@@ -160,11 +160,6 @@ func typecheck(n ir.Node, top int) (res ir.Node) {
 	lno := ir.SetPos(n)
 	defer func() { base.Pos = lno }()
 
-	// Skip over parens.
-	for n.Op() == ir.OPAREN {
-		n = n.(*ir.ParenExpr).X
-	}
-
 	// Skip typecheck if already done.
 	// But re-typecheck ONAME/OTYPE/OLITERAL/OPACK node in case context has changed.
 	if n.Typecheck() == 1 || n.Typecheck() == 3 {
@@ -216,6 +211,11 @@ func indexlit(n ir.Node) ir.Node {
 
 // typecheck1 should ONLY be called from typecheck.
 func typecheck1(n ir.Node, top int) ir.Node {
+	// Skip over parens.
+	for n.Op() == ir.OPAREN {
+		n = n.(*ir.ParenExpr).X
+	}
+
 	switch n.Op() {
 	default:
 		ir.Dump("typecheck", n)
@@ -663,14 +663,14 @@ func RewriteMultiValueCall(n ir.InitNode, call ir.Node) {
 
 	switch n := n.(type) {
 	default:
-		base.Fatalf("rewriteMultiValueCall %+v", n.Op())
+		base.Fatalf("RewriteMultiValueCall %+v", n.Op())
 	case *ir.CallExpr:
 		n.Args = list
 	case *ir.ReturnStmt:
 		n.Results = list
 	case *ir.AssignListStmt:
 		if n.Op() != ir.OAS2FUNC {
-			base.Fatalf("rewriteMultiValueCall: invalid op %v", n.Op())
+			base.Fatalf("RewriteMultiValueCall: invalid op %v", n.Op())
 		}
 		as.SetOp(ir.OAS2FUNC)
 		n.SetOp(ir.OAS2)
