@@ -65,7 +65,7 @@ func (check *Checker) usage(scope *Scope) {
 		return cmpPos(unused[i].pos, unused[j].pos) < 0
 	})
 	for _, v := range unused {
-		check.softErrorf(v, UnusedVar, "%s declared and not used", v.name)
+		check.softErrorf(v, UnusedVar, "declared and not used: %s", v.name)
 	}
 
 	for _, scope := range scope.children {
@@ -183,7 +183,7 @@ func (check *Checker) suspendedCall(keyword string, call *ast.CallExpr) {
 	case statement:
 		return
 	default:
-		unreachable()
+		panic("unreachable")
 	}
 	check.errorf(&x, code, "%s %s %s", keyword, msg, &x)
 }
@@ -257,8 +257,10 @@ L:
 			// (quadratic algorithm, but these lists tend to be very short)
 			for _, vt := range seen[val] {
 				if Identical(v.typ, vt.typ) {
-					check.errorf(&v, DuplicateCase, "duplicate case %s in expression switch", &v)
-					check.error(atPos(vt.pos), DuplicateCase, "\tprevious case") // secondary error, \t indented
+					err := check.newError(DuplicateCase)
+					err.addf(&v, "duplicate case %s in expression switch", &v)
+					err.addf(atPos(vt.pos), "previous case")
+					err.report()
 					continue L
 				}
 			}
@@ -270,14 +272,18 @@ L:
 // isNil reports whether the expression e denotes the predeclared value nil.
 func (check *Checker) isNil(e ast.Expr) bool {
 	// The only way to express the nil value is by literally writing nil (possibly in parentheses).
-	if name, _ := unparen(e).(*ast.Ident); name != nil {
+	if name, _ := ast.Unparen(e).(*ast.Ident); name != nil {
 		_, ok := check.lookup(name.Name).(*Nil)
 		return ok
 	}
 	return false
 }
 
-// If the type switch expression is invalid, x is nil.
+// caseTypes typechecks the type expressions of a type case, checks for duplicate types
+// using the seen map, and verifies that each type is valid with respect to the type of
+// the operand x in the type switch clause. If the type switch expression is invalid, x
+// must be nil. The result is the type of the last type expression; it is nil if the
+// expression denotes the predeclared nil.
 func (check *Checker) caseTypes(x *operand, types []ast.Expr, seen map[Type]ast.Expr) (T Type) {
 	var dummy operand
 L:
@@ -301,8 +307,10 @@ L:
 				if T != nil {
 					Ts = TypeString(T, check.qualifier)
 				}
-				check.errorf(e, DuplicateCase, "duplicate case %s in type switch", Ts)
-				check.error(other, DuplicateCase, "\tprevious case") // secondary error, \t indented
+				err := check.newError(DuplicateCase)
+				err.addf(e, "duplicate case %s in type switch", Ts)
+				err.addf(other, "previous case")
+				err.report()
 				continue L
 			}
 		}
@@ -341,11 +349,10 @@ L:
 // 			if T != nil {
 // 				Ts = TypeString(T, check.qualifier)
 // 			}
-// 			var err error_
-//			err.code = DuplicateCase
-// 			err.errorf(e, "duplicate case %s in type switch", Ts)
-// 			err.errorf(other, "previous case")
-// 			check.report(&err)
+// 			err := check.newError(_DuplicateCase)
+// 			err.addf(e, "duplicate case %s in type switch", Ts)
+// 			err.addf(other, "previous case")
+// 			err.report()
 // 			continue L
 // 		}
 // 		seen[hash] = e
@@ -511,8 +518,10 @@ func (check *Checker) stmt(ctxt stmtContext, s ast.Stmt) {
 			// with the same name as a result parameter is in scope at the place of the return."
 			for _, obj := range res.vars {
 				if alt := check.lookup(obj.name); alt != nil && alt != obj {
-					check.errorf(s, OutOfScopeResult, "result parameter %s not in scope at return", obj.name)
-					check.errorf(alt, OutOfScopeResult, "\tinner declaration of %s", obj)
+					err := check.newError(OutOfScopeResult)
+					err.addf(s, "result parameter %s not in scope at return", obj.name)
+					err.addf(alt, "inner declaration of %s", obj)
+					err.report()
 					// ok to continue
 				}
 			}
@@ -682,20 +691,19 @@ func (check *Checker) stmt(ctxt stmtContext, s ast.Stmt) {
 			check.error(s, InvalidSyntaxTree, "incorrect form of type switch guard")
 			return
 		}
-		var x operand
-		check.expr(nil, &x, expr.X)
-		if x.mode == invalid {
-			return
-		}
-		// TODO(gri) we may want to permit type switches on type parameter values at some point
+
 		var sx *operand // switch expression against which cases are compared against; nil if invalid
-		if isTypeParam(x.typ) {
-			check.errorf(&x, InvalidTypeSwitch, "cannot use type switch on type parameter value %s", &x)
-		} else {
-			if _, ok := under(x.typ).(*Interface); ok {
-				sx = &x
-			} else {
-				check.errorf(&x, InvalidTypeSwitch, "%s is not an interface", &x)
+		{
+			var x operand
+			check.expr(nil, &x, expr.X)
+			if x.mode != invalid {
+				if isTypeParam(x.typ) {
+					check.errorf(&x, InvalidTypeSwitch, "cannot use type switch on type parameter value %s", &x)
+				} else if IsInterface(x.typ) {
+					sx = &x
+				} else {
+					check.errorf(&x, InvalidTypeSwitch, "%s is not an interface", &x)
+				}
 			}
 		}
 
@@ -720,7 +728,10 @@ func (check *Checker) stmt(ctxt stmtContext, s ast.Stmt) {
 				// exactly one type, the variable has that type; otherwise, the variable
 				// has the type of the expression in the TypeSwitchGuard."
 				if len(clause.List) != 1 || T == nil {
-					T = x.typ
+					T = Typ[Invalid]
+					if sx != nil {
+						T = sx.typ
+					}
 				}
 				obj := NewVar(lhs.Pos(), check.pkg, lhs.Name, T)
 				scopePos := clause.Pos() + token.Pos(len("default")) // for default clause (len(List) == 0)
@@ -779,7 +790,7 @@ func (check *Checker) stmt(ctxt stmtContext, s ast.Stmt) {
 
 			// if present, rhs must be a receive operation
 			if rhs != nil {
-				if x, _ := unparen(rhs).(*ast.UnaryExpr); x != nil && x.Op == token.ARROW {
+				if x, _ := ast.Unparen(rhs).(*ast.UnaryExpr); x != nil && x.Op == token.ARROW {
 					valid = true
 				}
 			}
@@ -837,7 +848,7 @@ func (check *Checker) rangeStmt(inner stmtContext, s *ast.RangeStmt) {
 	type identType = ast.Ident
 	identName := func(n *identType) string { return n.Name }
 	sKey, sValue := s.Key, s.Value
-	var sExtra ast.Expr = nil
+	var sExtra ast.Expr = nil // (used only in types2 fork)
 	isDef := s.Tok == token.DEFINE
 	rangeVar := s.X
 	noNewVarPos := inNode(s, s.TokPos)
@@ -852,8 +863,8 @@ func (check *Checker) rangeStmt(inner stmtContext, s *ast.RangeStmt) {
 	var key, val Type
 	if x.mode != invalid {
 		// Ranging over a type parameter is permitted if it has a core type.
-		k, v, cause, isFunc, ok := rangeKeyVal(x.typ, func(v goVersion) bool {
-			return check.allowVersion(check.pkg, x.expr, v)
+		k, v, cause, ok := rangeKeyVal(x.typ, func(v goVersion) bool {
+			return check.allowVersion(x.expr, v)
 		})
 		switch {
 		case !ok && cause != "":
@@ -866,17 +877,6 @@ func (check *Checker) rangeStmt(inner stmtContext, s *ast.RangeStmt) {
 			check.softErrorf(sValue, InvalidIterVar, "range over %s permits only one iteration variable", &x)
 		case sExtra != nil:
 			check.softErrorf(sExtra, InvalidIterVar, "range clause permits at most two iteration variables")
-		case isFunc && ((k == nil) != (sKey == nil) || (v == nil) != (sValue == nil)):
-			var count string
-			switch {
-			case k == nil:
-				count = "no iteration variables"
-			case v == nil:
-				count = "one iteration variable"
-			default:
-				count = "two iteration variables"
-			}
-			check.softErrorf(&x, InvalidIterVar, "range over %s must have %s", &x, count)
 		}
 		key, val = k, v
 	}
@@ -893,7 +893,7 @@ func (check *Checker) rangeStmt(inner stmtContext, s *ast.RangeStmt) {
 	lhs := [2]Expr{sKey, sValue} // sKey, sValue may be nil
 	rhs := [2]Type{key, val}     // key, val may be nil
 
-	constIntRange := x.mode == constant_ && isInteger(x.typ)
+	rangeOverInt := isInteger(x.typ)
 
 	if isDef {
 		// short variable declaration
@@ -918,19 +918,28 @@ func (check *Checker) rangeStmt(inner stmtContext, s *ast.RangeStmt) {
 				check.errorf(lhs, InvalidSyntaxTree, "cannot declare %s", lhs)
 				obj = NewVar(lhs.Pos(), check.pkg, "_", nil) // dummy variable
 			}
+			assert(obj.typ == nil)
 
-			// initialize lhs variable
-			if constIntRange {
-				check.initVar(obj, &x, "range clause")
-			} else if typ := rhs[i]; typ != nil {
-				x.mode = value
-				x.expr = lhs // we don't have a better rhs expression to use here
-				x.typ = typ
-				check.initVar(obj, &x, "assignment") // error is on variable, use "assignment" not "range clause"
-			} else {
+			// initialize lhs iteration variable, if any
+			typ := rhs[i]
+			if typ == nil || typ == Typ[Invalid] {
+				// typ == Typ[Invalid] can happen if allowVersion fails.
 				obj.typ = Typ[Invalid]
 				obj.used = true // don't complain about unused variable
+				continue
 			}
+
+			if rangeOverInt {
+				assert(i == 0) // at most one iteration variable (rhs[1] == nil or Typ[Invalid] for rangeOverInt)
+				check.initVar(obj, &x, "range clause")
+			} else {
+				var y operand
+				y.mode = value
+				y.expr = lhs // we don't have a better rhs expression to use here
+				y.typ = typ
+				check.initVar(obj, &y, "assignment") // error is on variable, use "assignment" not "range clause"
+			}
+			assert(obj.typ != nil)
 		}
 
 		// declare variables
@@ -949,21 +958,36 @@ func (check *Checker) rangeStmt(inner stmtContext, s *ast.RangeStmt) {
 				continue
 			}
 
-			if constIntRange {
+			// assign to lhs iteration variable, if any
+			typ := rhs[i]
+			if typ == nil || typ == Typ[Invalid] {
+				continue
+			}
+
+			if rangeOverInt {
+				assert(i == 0) // at most one iteration variable (rhs[1] == nil or Typ[Invalid] for rangeOverInt)
 				check.assignVar(lhs, nil, &x, "range clause")
-			} else if typ := rhs[i]; typ != nil {
-				x.mode = value
-				x.expr = lhs // we don't have a better rhs expression to use here
-				x.typ = typ
-				check.assignVar(lhs, nil, &x, "assignment") // error is on variable, use "assignment" not "range clause"
+				// If the assignment succeeded, if x was untyped before, it now
+				// has a type inferred via the assignment. It must be an integer.
+				// (go.dev/issues/67027)
+				if x.mode != invalid && !isInteger(x.typ) {
+					check.softErrorf(lhs, InvalidRangeExpr, "cannot use iteration variable of type %s", x.typ)
+				}
+			} else {
+				var y operand
+				y.mode = value
+				y.expr = lhs // we don't have a better rhs expression to use here
+				y.typ = typ
+				check.assignVar(lhs, nil, &y, "assignment") // error is on variable, use "assignment" not "range clause"
 			}
 		}
-	} else if constIntRange {
+	} else if rangeOverInt {
 		// If we don't have any iteration variables, we still need to
 		// check that a (possibly untyped) integer range expression x
 		// is valid.
 		// We do this by checking the assignment _ = x. This ensures
-		// that an untyped x can be converted to a value of type int.
+		// that an untyped x can be converted to a value of its default
+		// type (rune or int).
 		check.assignment(&x, nil, "range clause")
 	}
 
@@ -975,9 +999,9 @@ func (check *Checker) rangeStmt(inner stmtContext, s *ast.RangeStmt) {
 // If allowVersion != nil, it is used to check the required language version.
 // If the range clause is not permitted, rangeKeyVal returns ok = false.
 // When ok = false, rangeKeyVal may also return a reason in cause.
-func rangeKeyVal(typ Type, allowVersion func(goVersion) bool) (key, val Type, cause string, isFunc, ok bool) {
-	bad := func(cause string) (Type, Type, string, bool, bool) {
-		return Typ[Invalid], Typ[Invalid], cause, false, false
+func rangeKeyVal(typ Type, allowVersion func(goVersion) bool) (key, val Type, cause string, ok bool) {
+	bad := func(cause string) (Type, Type, string, bool) {
+		return Typ[Invalid], Typ[Invalid], cause, false
 	}
 	toSig := func(t Type) *Signature {
 		sig, _ := coreType(t).(*Signature)
@@ -990,29 +1014,28 @@ func rangeKeyVal(typ Type, allowVersion func(goVersion) bool) (key, val Type, ca
 		return bad("no core type")
 	case *Basic:
 		if isString(typ) {
-			return Typ[Int], universeRune, "", false, true // use 'rune' name
+			return Typ[Int], universeRune, "", true // use 'rune' name
 		}
 		if isInteger(typ) {
 			if allowVersion != nil && !allowVersion(go1_22) {
 				return bad("requires go1.22 or later")
 			}
-			return orig, nil, "", false, true
+			return orig, nil, "", true
 		}
 	case *Array:
-		return Typ[Int], typ.elem, "", false, true
+		return Typ[Int], typ.elem, "", true
 	case *Slice:
-		return Typ[Int], typ.elem, "", false, true
+		return Typ[Int], typ.elem, "", true
 	case *Map:
-		return typ.key, typ.elem, "", false, true
+		return typ.key, typ.elem, "", true
 	case *Chan:
 		if typ.dir == SendOnly {
 			return bad("receive from send-only channel")
 		}
-		return typ.elem, nil, "", false, true
+		return typ.elem, nil, "", true
 	case *Signature:
-		// TODO(gri) when this becomes enabled permanently, add version check
-		if !buildcfg.Experiment.RangeFunc {
-			break
+		if !buildcfg.Experiment.RangeFunc && allowVersion != nil && !allowVersion(go1_23) {
+			return bad("requires go1.23 or later")
 		}
 		assert(typ.Recv() == nil)
 		switch {
@@ -1037,7 +1060,7 @@ func rangeKeyVal(typ Type, allowVersion func(goVersion) bool) (key, val Type, ca
 		if cb.Params().Len() >= 2 {
 			val = cb.Params().At(1).Type()
 		}
-		return key, val, "", true, true
+		return key, val, "", true
 	}
 	return
 }
